from itertools import combinations


def dump_hypodd(families, catalogue, errors, repeaters, parameters, output_dir):
    for i, family in enumerate(families):
        with (output_dir / f"{i}_RES_event.sel").open("w") as file:
            for n in family:
                e_t, e_h, e_v = map(lambda s: err if (err := errors[catalogue.loc[n, "name"]].get(s)) else parameters[
                    'hypodd_default_' + s], ['time_uncertainty', 'horizontal_uncertainty', 'vertical_uncertainty'])
                file.write(f"{catalogue.loc[n, 'date'].strftime('%Y%m%d  %H%M%S%f')[:-4]}   "
                           f"{catalogue.loc[n, 'latitude']:.4f}     {catalogue.loc[n, 'longitude']:.4f}    "
                           f"{catalogue.loc[n, 'depth']:.3f}   {catalogue.loc[n, 'magnitude']:.1f}    "
                           f"{e_t:.2f}    {e_h:.2f}   {e_v:.2f}        "
                           f"{catalogue.loc[n, 'name']}\n")
        if len(family) > 2:
            with (output_dir / f"{i}_RES_dt.cc").open("w") as file:
                for (t1, t2) in combinations(family, 2):
                    if repeaters[(t1, t2)]:
                        file.write(f"#    {catalogue.loc[t1, 'name']}    {catalogue.loc[t2, 'name']}     0.0\n")
                        for (network, station) in sorted(repeaters[(t1, t2)]):
                            cc, delta_sp = repeaters[(t1, t2)][(network, station)]
                            delta_v = parameters['Vp'] - parameters['Vs']
                            ttp = -parameters['Vs'] * delta_sp / delta_v
                            file.write(f"{station}     {ttp: 10.9f}    {cc:.2f}    P\n")
                            tts = parameters['Vp'] * delta_sp / delta_v
                            file.write(f"{station}     {tts: 10.9f}    {cc:.2f}    S\n")



from obspy import UTCDateTime


def _rise_phases(origin_time, station, phase_file):
    p_pick, s_pick = None, None
    with phase_file.open('r') as file:
        start_section = False
        event_origin = None
        while line := file.readline():
            if line.isspace():
                if start_section:
                    break
                else:
                    continue
            fields = line.split()
            if len(fields) == 12:
                event_origin = fields[2] + fields[3] + fields[4] + fields[7] + fields[8] + fields[9]
            if event_origin and event_origin.startswith(origin_time.strftime('%Y%m%d%H%M%S')):
                start_section = True
                pick_datetime = fields[6] + fields[7] + fields[8]
                pick_station = fields[0]
                pick_type = fields[4]
                if pick_station == station:
                    if pick_type == 'P':
                        p_pick = UTCDateTime.strptime(pick_datetime, '%Y%m%d%H%M%S.%f')
                    elif pick_type == 'S':
                        s_pick = UTCDateTime.strptime(pick_datetime, '%Y%m%d%H%M%S.%f')

    return p_pick, s_pick


def _rise_errors(origin_time, phase_file):
    return {}


def _hyposynth_phases(origin_time, station, phase_file):
    with phase_file.open('r') as file:
        lines = []
        start_section = False
        while line := file.readline():
            if line.startswith(origin_time.strftime('%Y%m%d%H%M%S')):
                start_section = True
                continue
            if start_section and line.startswith(30 * " "):
                break
            if start_section:
                lines.append(line)
    p_pick, s_pick = None, None
    for line in lines:
        if line.startswith(station) and int(line[16]) < 4:
            if not line[31:39].isspace():
                p_pick = UTCDateTime.strptime(line[17:38], '%Y%m%d%H%M %S%f')
            if not line[43:51].isspace():
                s_pick = UTCDateTime.strptime(line[17:30] + line[42:50], '%Y%m%d%H%M %S%f')
    return p_pick, s_pick


def _hyposynth_errors(origin_time, phase_file):
    return {}


registry = {'rise_custom': {'phases': _rise_phases, 'errors': _rise_errors},
            'hyposynth': {'phases': _hyposynth_phases, 'errors': _hyposynth_errors}}



import numpy as np
from obspy.signal.cross_correlation import correlate_template

from findres.utils import longest_subsequence_of_trues


class EmptyOrConstantTraceError(BaseException):
    pass


class LowCoherenceError(BaseException):
    pass


def correlate_waves(trace1, trace2, max_shift, **kwargs):
    if trace1.stats.npts == 0 or trace1.stats.npts == 0:
        raise EmptyOrConstantTraceError("Cannot cross-correlate empty traces.")
    elif trace1.std() == 0 or trace2.std() == 0:
        raise EmptyOrConstantTraceError("Normalized cross-correlation of constant traces is undefined.")
    if trace1.stats.npts >= trace2.stats.npts:
        is_swapped = False
        data1, data2 = trace1.data, trace2.data
    else:
        is_swapped = True
        data1, data2 = trace2.data, trace1.data
    data1 = np.concatenate([np.zeros(max_shift), data1, np.zeros(max_shift)])
    cc = correlate_template(data1, data2, **kwargs)
    argmax = np.argmax(cc)
    shift = argmax - max_shift
    if is_swapped:
        shift = -shift
    return shift, cc[argmax]


def find_slope(cross_spectrum, frequency_range, params):
    coherence = cross_spectrum.cohe.flatten()
    frequency = cross_spectrum.freq.flatten()
    phase = cross_spectrum.phase.flatten()
    lower_frequency_bound = frequency_range[0],
    upper_frequency_bound = frequency_range[1] + params['coherence_frequency_upper_bound_correction']

    indices = longest_subsequence_of_trues((coherence > params['cs_coherence_threshold']) &
                                           (frequency > lower_frequency_bound) & (frequency < upper_frequency_bound))
    if indices:
        start, stop = indices
        if (stop - start) < params['coherence_minimum_num_points']:
            raise LowCoherenceError("Not enough points above coherence threshold.")
        else:
            x = frequency[start:stop]
            y = phase[start:stop] * np.pi / 180.0
            slope = np.mean(x * y) / np.mean(x * x)
            return slope, indices
    else:
        raise LowCoherenceError("No points above coherence threshold.")
